@@ -22,6 +22,16 @@
 
 #include <gio/gio.h>
 
+#include "request.h"
+#include "permissions.h"
+
+Permission device_get_permission_sync (const char *app_id,
+                                       const char *device);
+
+gboolean device_query_permission_sync (const char *app_id,
+                                       const char *device,
+                                       const char *request_handle);
+
 GDBusInterfaceSkeleton * device_create (GDBusConnection *connection,
                                         const char      *dbus_name,
                                         gpointer         lockdown);
