@@ -23,4 +23,5 @@
 #include <gio/gio.h>
 
 GDBusInterfaceSkeleton * print_create (GDBusConnection *connection,
-                                       const char      *dbus_name);
+                                       const char      *dbus_name,
+                                       gpointer         lockdown);
