@@ -290,6 +290,26 @@ get_token (GDBusMethodInvocation *invocation)
     {
       // no request objects
     }
+  else if (strcmp (interface, "org.freedesktop.portal.GameMode") == 0)
+    {
+      // no request objects
+    }
+  else if (strcmp (interface, "org.freedesktop.portal.Camera") == 0)
+    {
+      if (strcmp (method, "AccessCamera") == 0 )
+        {
+          options = g_variant_get_child_value (parameters, 0);
+        }
+      else if (strcmp (method, "OpenPipewireRemote") == 0)
+        {
+          // no request objects
+        }
+      else
+        {
+          g_warning ("Support for %s::%s missing in %s",
+                     interface, method, G_STRLOC);
+        }
+    }
   else
     {
       g_print ("Support for %s missing in " G_STRLOC, interface);
