@@ -32,6 +32,7 @@
 #include "xdp-impl-dbus.h"
 #include "request.h"
 #include "call.h"
+#include "portal-impl.h"
 #include "documents.h"
 #include "permissions.h"
 #include "file-chooser.h"
@@ -50,6 +51,9 @@
 #include "trash.h"
 #include "location.h"
 #include "settings.h"
+#include "background.h"
+#include "gamemode.h"
+#include "camera.h"
 
 static GMainLoop *loop = NULL;
 
@@ -91,197 +95,6 @@ printerr_handler (const gchar *string)
   fprintf (stderr, "%serror: %s%s\n", prefix, suffix, string);
 }
 
-typedef struct {
-  char *source;
-  char *dbus_name;
-  char **interfaces;
-  char **use_in;
-  int priority;
-} PortalImplementation;
-
-static void
-portal_implementation_free (PortalImplementation *impl)
-{
-  g_free (impl->source);
-  g_free (impl->dbus_name);
-  g_strfreev (impl->interfaces);
-  g_strfreev (impl->use_in);
-  g_free (impl);
-}
-
-G_DEFINE_AUTOPTR_CLEANUP_FUNC(PortalImplementation, portal_implementation_free)
-
-static GList *implementations = NULL;
-
-static gboolean
-register_portal (const char *path, GError **error)
-{
-  g_autoptr(GKeyFile) keyfile = g_key_file_new ();
-  g_autoptr(PortalImplementation) impl = g_new0 (PortalImplementation, 1);
-  int i;
-
-  g_debug ("loading %s", path);
-
-  if (!g_key_file_load_from_file (keyfile, path, G_KEY_FILE_NONE, error))
-    return FALSE;
-
-  impl->source = g_path_get_basename (path);
-  impl->dbus_name = g_key_file_get_string (keyfile, "portal", "DBusName", error);
-  if (impl->dbus_name == NULL)
-    return FALSE;
-  if (!g_dbus_is_name (impl->dbus_name))
-    {
-      g_set_error (error, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_INVALID_VALUE,
-                   "Not a valid bus name: %s", impl->dbus_name);
-      return FALSE;
-    }
-
-  impl->interfaces = g_key_file_get_string_list (keyfile, "portal", "Interfaces", NULL, error);
-  if (impl->interfaces == NULL)
-    return FALSE;
-  for (i = 0; impl->interfaces[i]; i++)
-    {
-      if (!g_dbus_is_interface_name (impl->interfaces[i]))
-        {
-          g_set_error (error, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_INVALID_VALUE,
-                       "Not a valid interface name: %s", impl->interfaces[i]);
-          return FALSE;
-        }
-      if (!g_str_has_prefix (impl->interfaces[i], "org.freedesktop.impl.portal."))
-        {
-          g_set_error (error, G_KEY_FILE_ERROR, G_KEY_FILE_ERROR_INVALID_VALUE,
-                       "Not a portal backend interface: %s", impl->interfaces[i]);
-          return FALSE;
-        }
-    }
-
-  impl->use_in = g_key_file_get_string_list (keyfile, "portal", "UseIn", NULL, error);
-  if (impl->use_in == NULL)
-    return FALSE;
-
-  if (opt_verbose)
-    {
-      g_autofree char *uses = g_strjoinv (", ", impl->use_in);
-      g_debug ("portal implementation for %s", uses);
-      for (i = 0; impl->interfaces[i]; i++)
-        g_debug ("portal implementation supports %s", impl->interfaces[i]);
-    }
-
-  implementations = g_list_prepend (implementations, impl);
-  impl = NULL;
-
-  return TRUE;
-}
-
-static gint
-sort_impl_by_name (gconstpointer a,
-                   gconstpointer b)
-{
-  const PortalImplementation *pa = a;
-  const PortalImplementation *pb = b;
-
-  return strcmp (pa->source, pb->source);
-}
-
-static void
-load_installed_portals (void)
-{
-  const char *portal_dir = PKGDATADIR "/portals";
-  g_autoptr(GFile) dir = g_file_new_for_path (portal_dir);
-  g_autoptr(GFileEnumerator) enumerator = NULL;
-
-  enumerator = g_file_enumerate_children (dir, "*", G_FILE_QUERY_INFO_NONE, NULL, NULL);
-
-  if (enumerator == NULL)
-    return;
-
-  while (TRUE)
-    {
-      g_autoptr(GFileInfo) info = g_file_enumerator_next_file (enumerator, NULL, NULL);
-      g_autoptr(GFile) child = NULL;
-      g_autofree char *path = NULL;
-      const char *name;
-      g_autoptr(GError) error = NULL;
-
-      if (info == NULL)
-        break;
-
-      name = g_file_info_get_name (info);
-
-      if (!g_str_has_suffix (name, ".portal"))
-        continue;
-
-      child = g_file_enumerator_get_child (enumerator, info);
-      path = g_file_get_path (child);
-
-      if (!register_portal (path, &error))
-        {
-          g_warning ("Error loading %s: %s", path, error->message);
-          continue;
-        }
-    }
-
-  implementations = g_list_sort (implementations, sort_impl_by_name);
-}
-
-static gboolean
-g_strv_case_contains (const gchar * const *strv,
-                      const gchar         *str)
-{
-  for (; *strv != NULL; strv++)
-    {
-      if (g_ascii_strcasecmp (str, *strv) == 0)
-        return TRUE;
-    }
-
-  return FALSE;
-}
-
-static PortalImplementation *
-find_portal_implementation (const char *interface)
-{
-  const char *desktops_str = g_getenv ("XDG_CURRENT_DESKTOP");
-  g_auto(GStrv) desktops = NULL;
-  int i;
-  GList *l;
-
-  if (desktops_str == NULL)
-    desktops_str = "";
-
-  desktops = g_strsplit (desktops_str, ":", -1);
-
-  for (i = 0; desktops[i] != NULL; i++)
-    {
-     for (l = implementations; l != NULL; l = l->next)
-        {
-          PortalImplementation *impl = l->data;
-
-          if (!g_strv_contains ((const char **)impl->interfaces, interface))
-            continue;
-
-          if (g_strv_case_contains ((const char **)impl->use_in, desktops[i]))
-            {
-              g_debug ("Using %s for %s in %s", impl->source, interface, desktops[i]);
-              return impl;
-            }
-        }
-    }
-
-  /* Fall back to *any* installed implementation */
-  for (l = implementations; l != NULL; l = l->next)
-    {
-      PortalImplementation *impl = l->data;
-
-      if (!g_strv_contains ((const char **)impl->interfaces, interface))
-        continue;
-
-      g_debug ("Falling back to %s for %s", impl->source, interface);
-      return impl;
-    }
-
-  return NULL;
-}
-
 static gboolean
 method_needs_request (GDBusMethodInvocation *invocation)
 {
@@ -305,6 +118,13 @@ method_needs_request (GDBusMethodInvocation *invocation)
       else
         return TRUE;
     }
+  if (strcmp (interface, "org.freedesktop.portal.Camera") == 0)
+    {
+      if (strcmp (method, "OpenPipeWireRemote") == 0)
+        return FALSE;
+      else
+        return TRUE;
+    }
   else
     {
       return TRUE;
@@ -380,6 +200,7 @@ on_bus_acquired (GDBusConnection *connection,
                  gpointer         user_data)
 {
   PortalImplementation *implementation;
+  PortalImplementation *implementation2;
   g_autoptr(GError) error = NULL;
   XdpImplLockdown *lockdown;
 
@@ -400,6 +221,7 @@ on_bus_acquired (GDBusConnection *connection,
   export_portal_implementation (connection, network_monitor_create (connection));
   export_portal_implementation (connection, proxy_resolver_create (connection));
   export_portal_implementation (connection, trash_create (connection));
+  export_portal_implementation (connection, game_mode_create (connection));
 
   implementation = find_portal_implementation ("org.freedesktop.impl.portal.Settings");
   export_portal_implementation (connection, settings_create (connection, implementation ? implementation->dbus_name : NULL));
@@ -435,6 +257,7 @@ on_bus_acquired (GDBusConnection *connection,
                                   inhibit_create (connection, implementation->dbus_name));
 
   implementation = find_portal_implementation ("org.freedesktop.impl.portal.Access");
+  implementation2 = find_portal_implementation ("org.freedesktop.impl.portal.Background");
   if (implementation != NULL)
     {
       export_portal_implementation (connection,
@@ -443,8 +266,18 @@ on_bus_acquired (GDBusConnection *connection,
       export_portal_implementation (connection,
                                     location_create (connection, implementation->dbus_name, lockdown));
 #endif
+
+#if HAVE_PIPEWIRE
+      export_portal_implementation (connection, camera_create (connection));
+#endif
     }
 
+  if (implementation != NULL && implementation2 != NULL)
+    export_portal_implementation (connection,
+                                  background_create (connection,
+                                                     implementation->dbus_name,
+                                                     implementation2->dbus_name));
+
   implementation = find_portal_implementation ("org.freedesktop.impl.portal.Account");
   if (implementation != NULL)
     export_portal_implementation (connection,
@@ -540,7 +373,7 @@ main (int argc, char *argv[])
 
   g_set_prgname (argv[0]);
 
-  load_installed_portals ();
+  load_installed_portals (opt_verbose);
 
   loop = g_main_loop_new (NULL, FALSE);
 
