@@ -35,6 +35,7 @@
 #define FLATPAK_METADATA_GROUP_INSTANCE "Instance"
 #define FLATPAK_METADATA_KEY_APP_PATH "app-path"
 #define FLATPAK_METADATA_KEY_RUNTIME_PATH "runtime-path"
+#define FLATPAK_METADATA_KEY_INSTANCE_ID "instance-id"
 
 gint xdp_mkstempat (int    dir_fd,
                     gchar *tmpl,
@@ -50,10 +51,15 @@ typedef struct _XdpAppInfo XdpAppInfo;
 XdpAppInfo *xdp_app_info_ref             (XdpAppInfo  *app_info);
 void        xdp_app_info_unref           (XdpAppInfo  *app_info);
 const char *xdp_app_info_get_id          (XdpAppInfo  *app_info);
+char *      xdp_app_info_get_instance    (XdpAppInfo  *app_info);
 gboolean    xdp_app_info_is_host         (XdpAppInfo  *app_info);
 gboolean    xdp_app_info_supports_opath  (XdpAppInfo  *app_info);
 char *      xdp_app_info_remap_path      (XdpAppInfo  *app_info,
                                           const char  *path);
+gboolean    xdg_app_info_map_pids        (XdpAppInfo  *app_info,
+					  pid_t       *pids,
+					  guint        n_pids,
+					  GError     **error);
 char *      xdp_app_info_get_path_for_fd (XdpAppInfo  *app_info,
                                           int          fd,
                                           int          require_st_mode,
@@ -62,6 +68,9 @@ char *      xdp_app_info_get_path_for_fd (XdpAppInfo  *app_info,
 gboolean    xdp_app_info_has_network     (XdpAppInfo  *app_info);
 XdpAppInfo *xdp_get_app_info_from_pid    (pid_t        pid,
                                           GError     **error);
+GAppInfo *  xdp_app_info_load_app_info   (XdpAppInfo *app_info);
+char **     xdp_app_info_rewrite_commandline (XdpAppInfo *app_info,
+                                              const char *const *commandline);
 
 G_DEFINE_AUTOPTR_CLEANUP_FUNC(XdpAppInfo, xdp_app_info_unref)
 
